/*
 * @(#)src/classes/sov/java/util/Hashtable.java, util, asdev, 20130203 1.28
 * ===========================================================================
 * Licensed Materials - Property of IBM
 * "Restricted Materials of IBM"
 *
 * IBM SDK, Java(tm) 2 Technology Edition, v5.0
 * (C) Copyright IBM Corp. 1998, 2005. All Rights Reserved
 * ===========================================================================
 */
 
 /*
 * Licensed Materials - Property of IBM,
 *     Copyright IBM Corp. 1998, 2009  All Rights Reserved
 */

/*
*
* Change activity:
*
* Reason Date   Origin   Description
* ------ ------ -------- ---------------------------------------------------
* 191994 290512 bhamaram add Hashtable in ClassLibraries from j9VM 
* 192023 300512 bhamaram LIR 26700:improve hashtable performace at high
*                         collision count
* 192811 040712 bhamaram Change the default jdk.map.althashing.threshold to -1
* 192925 130712 sakandur  SVT regression issue
* 193244 280812 sakandur 5% iLog performance regression due to LIR 26700
* ===========================================================================
*/

package java.util;


import java.io.Serializable;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.security.AccessController;
import java.security.PrivilegedAction;

/**
 * Hashtable associates keys with values. Keys and values cannot be null.
 * The size of the Hashtable is the number of key/value pairs it contains.
 * The capacity is the number of key/value pairs the Hashtable can hold.
 * The load factor is a float value which determines how full the Hashtable
 * gets before expanding the capacity. If the load factor of the Hashtable
 * is exceeded, the capacity is doubled.
 *
 * @author		OTI
 * @version		initial
 *
 * @see			Enumeration
 * @see			java.io.Serializable
 * @see			java.lang.Object#equals
 * @see			java.lang.Object#hashCode
 */

public class Hashtable<K, V> extends Dictionary<K, V> implements Map<K, V>, Cloneable, Serializable {
	private static final long serialVersionUID = 1421746759512286392L;
	private static final String initialPropValue = new String();
	private static String CLONE_HASHTABLE_FOR_SYNCHRONIZATION = initialPropValue;
	transient int elementCount;
	transient Entry<K, V>[] elementData;
	private float loadFactor;
	private int threshold;
	transient int firstSlot = 0;
	transient int lastSlot = -1;
	transient int modCount = 0;
	private static final Enumeration emptyEnumerator = new EmptyEnumerator();
	private static final Iterator emptyIterator = new EmptyIterator();

   
    

    /*ibm@192023 Start*/
	
	/**
     * The maximum size of array to allocate.
     * Some VMs reserve some header words in an array.
     * Attempts to allocate larger arrays may result in
     * OutOfMemoryError: Requested array size exceeds VM limit
     */
    private static final int MAX_ARRAY_SIZE = Integer.MAX_VALUE - 8;
	
	/**
     * The default threshold of capacity above which alternate hashing is used. 
     * Alternative hashing reduces the incidence of collisions due to weak hash
     * code calculation. 
     * <p/>
     * This value may be overridden by defining the system property 
     * {@code java.util.althashing.threshold}. A property value of {@code 1}
     * forces alternative hashing to be used at all times whereas
     * {@code 2147483648 } ({@code Integer.MAX_VALUE}) value ensures that
     * alternative hashing is never used.
     */
    static final int ALTERNATE_HASHING_THRESHOLD_DEFAULT = -1; /*ibm@192811*/
    
    private static class Holder {
        /**
         * Table capacity above which to switch to use alternate hashing.
         */
        static final int ALTERNATE_HASHING_THRESHOLD;
        
        static {
            String altThreshold = (String) java.security.AccessController.doPrivileged(
                new sun.security.action.GetPropertyAction(
                    "jdk.map.althashing.threshold"));
            
            int threshold;
            try {
                threshold = (null != altThreshold)
                        ? Integer.parseInt(altThreshold)
                        : ALTERNATE_HASHING_THRESHOLD_DEFAULT;
                        
                if(threshold == -1) {
                    threshold = Integer.MAX_VALUE;
                }
                
                if(threshold < 0) {
                    throw new IllegalArgumentException("value must be positive integer.");
                }
            } catch(IllegalArgumentException failed) {
                throw new Error("Illegal value for 'jdk.map.althashing.threshold'", failed);
            }
            
            ALTERNATE_HASHING_THRESHOLD = threshold;
        }
    }
            
    /** 
     * If {@code true} then perform alternate hashing to reduce the incidence of
     * collisions due to weak hash code calculation.
     */
    transient boolean useAltHashing;
    
    // Unsafe mechanics
    private static final sun.misc.Unsafe UNSAFE;
    private static final long HASHMASK_OFFSET;
    
     static {
        try {
            UNSAFE = sun.misc.Unsafe.getUnsafe();
            HASHMASK_OFFSET = UNSAFE.objectFieldOffset(
                Hashtable.class.getDeclaredField("hashMask"));                        
        } catch (Exception e) {
            throw new Error(e);
        }
     }
     
    /**
     * A random mask value that is used for hashcode values associated with this
     * instance to make hash collisions harder to find.
     */
    transient final int hashMask = sun.misc.VM.isBooted() && Holder.ALTERNATE_HASHING_THRESHOLD == Integer.MAX_VALUE ? 
            0 :sun.misc.Hashing.randomHashSeed(this); /*ibm@193244*/
    
    private int hash(Object k) {
        if (useAltHashing) {
            if (k instanceof String) {
                return sun.misc.Hashing.stringHash32((String) k);
            } else {
                int h = hashMask ^ k.hashCode();
                
                // This function ensures that hashCodes that differ only by
                // constant multiples at each bit position have a bounded
                // number of collisions (approximately 8 at default load factor).
                h ^= (h >>> 20) ^ (h >>> 12);
                return h ^ (h >>> 7) ^ (h >>> 4);
            }
        } else  {
            return k.hashCode();
        }
    }
    /*ibm@192023 End*/
	
    private static class EmptyEnumerator implements Enumeration {
    	public boolean hasMoreElements() {
    		return false;
    	}
    	public Object nextElement() {
    		throw new NoSuchElementException();
    	}
    }
    
    private static class EmptyIterator implements Iterator {
    	public boolean hasNext() {
    		return false;
    	}
    	public Object next() {
    		throw new NoSuchElementException();
    	}
    	public void remove() {
    		throw new IllegalStateException();
    	}
    }
    
    
    private static class Entry<K, V> extends MapEntry<K, V> {
    	Entry<K, V> next;
		int hashCode; /*ibm@192023*/
    
    	Entry(K theKey, V theValue, int hash) {
    		super(theKey, theValue);
			this.hashCode = hash; /*ibm@192023*/
    	}
    
    	public Object clone() {
    		Entry entry = (Entry)super.clone();
    		if (next != null) entry.next = (Entry)next.clone();
    		return entry;
    	}
    
    	public V setValue(V object) {
    		if (object == null) throw new NullPointerException();
    		V result = value;
    		value = object;
    		return result;
    	}
		
    
    	public boolean equalsKey(Object aKey, int hash) {
    		return hash == hashCode && key.equals(aKey); /*ibm@192023*/
    	}
    
    	public String toString() {
    		return key + "=" + value;
    	}
    }
    
    private final class HashIterator implements Iterator, Enumeration {
    	private int position, expectedModCount;
    	private MapEntry.Type type;
    	private Entry entry, removeEntry;
    	private int removePosition;
    
    HashIterator(MapEntry.Type value) {
    	type = value;
    	position = lastSlot + 1;
    	expectedModCount = modCount;
    }
    
    HashIterator(boolean isKey) {
    	type = isKey ? MapEntry.Type.keyEnumeration : MapEntry.Type.valueEnumeration;
    	position = lastSlot + 1;
    	expectedModCount = modCount;
    }
    
    public boolean hasMoreElements() {
    	if (entry != null) return true;
    	while (position > firstSlot)
    		if ((elementData[--position]) != null) {
    			entry = elementData[position];
    			return true;
    		}
    	return false;
    }
    
    public boolean hasNext() {
    	if (entry != null) return true;
    	while (position > firstSlot)
    		if ((entry = elementData[--position]) != null) {
    			return true;
    		}
    	return false;
    }
    
    public Object nextElement() {
    	if (hasNext()) {
    		removeEntry = entry;
    		removePosition = position;
    		Object result = type.get(entry);
    		entry = entry.next;
    		return result;
    	}
    	throw new NoSuchElementException();
    }
    
    public Object next() {
    	if (expectedModCount == modCount) {
    		if (hasNext()) {
    			removeEntry = entry;
    			removePosition = position;
    			Object result = type.get(entry);
    			entry = entry.next;
    			return result;
    		}
    		throw new NoSuchElementException();
    	} else throw new ConcurrentModificationException();
    }
    
    public void remove() {
    	if (type.isIterator()) {
    		if (expectedModCount == modCount) {
    			if (removeEntry != null) {
    				synchronized(Hashtable.this) {
    					boolean removed = false;
    					Entry entry = elementData[removePosition];
    					if (entry == removeEntry) {
    						elementData[removePosition] = entry.next;
    						removed = true;
    					} else {
    						while (entry != null && entry.next != removeEntry) {
    							entry = entry.next;
    						}
    						if (entry != null) {
    							entry.next = removeEntry.next;
    							removed = true;
    						}
    					}
    					removeEntry = null;
    					if (removed) {
    						modCount++;
    						elementCount--;
    						expectedModCount++;
    						return;
    					}
    					// the entry must have been (re)moved outside of the iterator
    					// but this condition wasn't caught by the modCount check
    					// throw ConcurrentModificationException() outside of synchronized block
    				}
    			} else throw new IllegalStateException();
    		}
    		throw new ConcurrentModificationException();
    	}
    	throw new UnsupportedOperationException();
    }
    }
    
    /**
     * Constructs a new Hashtable using the default capacity
     * and load factor.
     */
    public Hashtable() {
    	this(11);
    }
    /**
     * Constructs a new Hashtable using the specified capacity
     * and the default load factor.
     *
     * @param		capacity	the initial capacity
     */
    public Hashtable(int capacity) {
        this(capacity,0.75f);             /*ibm@192023*/
    }
    
    /**
     * Constructs a new Hashtable using the specified capacity
     * and load factor.
     *
     * @param		capacity	the initial capacity
     * @param		loadFactor	the initial load factor
     */
    public Hashtable(int capacity, float loadFactor) {
    	if (capacity >= 0 && loadFactor > 0) {
    		elementCount = 0;
            if(capacity == 0)  /*ibm@192023*/
                capacity = 1;  /*ibm@192023*/
			firstSlot = capacity;
    		elementData = new Entry[capacity]; /*ibm@192023*/
    		this.loadFactor = loadFactor;
    		computeMaxSize();
            useAltHashing = sun.misc.VM.isBooted() &&
                (capacity >= Holder.ALTERNATE_HASHING_THRESHOLD); /*ibm@192023*/
    	} else throw new IllegalArgumentException();
    }
    
    /**
     * Constructs a new instance of Hashtable containing the mappings from the
     * specified Map.
     *
     * @param		map	the mappings to add
     */
    public Hashtable(Map<? extends K, ? extends V> map) {
    	this(map.size() < 6 ? 11 : (map.size() * 4 / 3) + 11);
    	putAll(map);
    }
    
    /**
     * Removes all key/value pairs from this Hashtable, leaving the size zero
     * and the capacity unchanged.
     *
     * @see			#isEmpty
     * @see			#size
     */
    public synchronized void clear() {
    	elementCount = 0;
    	Arrays.fill(elementData, null);
    	modCount++;
    }
    
    /**
     * Answers a new Hashtable with the same key/value pairs, capacity
     * and load factor.
     *
     * @return		a shallow copy of this Hashtable
     *
     * @see			java.lang.Cloneable
     */
    public synchronized Object clone() {
    	try {
    		Hashtable hashtable = (Hashtable) super.clone();
    		hashtable.elementData = new Entry[elementData.length];
    		Entry entry;
    		for (int i=elementData.length; --i >= 0;)
    			if ((entry = elementData[i]) != null)
    				hashtable.elementData[i] = (Entry)entry.clone();
    		return hashtable;
    	} catch (CloneNotSupportedException e) {return null;}
    }
    
    private void computeMaxSize() {
		threshold = (int)Math.min(elementData.length * loadFactor, MAX_ARRAY_SIZE + 1); /*ibm@192023*/
    }
    /**
     * Answers if this Hashtable contains the specified object as the value
     * of at least one of the key/value pairs.
     *
     * @param		value	the object to look for as a value in this Hashtable
     * @return		true if object is a value in this Hashtable, false otherwise
     *
     * @see			#containsKey
     * @see			java.lang.Object#equals
     */
    public synchronized boolean contains(Object value) {
    	if (value != null) {
    		for (int i=elementData.length; --i >= 0;) {
    			Entry entry = elementData[i];
    			while (entry != null) {
    				if (value.equals(entry.value)) return true;
    				entry = entry.next;
    			}
    		}
    		return false;
    	} else throw new NullPointerException();
    }
    /**
     * Answers if this Hashtable contains the specified object as a key
     * of one of the key/value pairs.
     *
     * @param		key	the object to look for as a key in this Hashtable
     * @return		true if object is a key in this Hashtable, false otherwise
     *
     * @see			#contains
     * @see			java.lang.Object#equals
     */
    public synchronized boolean containsKey(Object key) {
    	return getEntry(key) != null;
    }
    
    /**
     * Searches this Hashtable for the specified value.
     *
     * @param		value the object to search for
     * @return		true if <code>value</code> is a value of this Hashtable, false otherwise
     */
    public boolean containsValue(Object value) {
    	return contains(value);
    }
    
    /**
     * Answers an Enumeration on the values of this Hashtable. The
     * results of the Enumeration may be affected if the contents
     * of this Hashtable are modified.
     *
     * @return		an Enumeration of the values of this Hashtable
     *
     * @see			#keys
     * @see			#size
     * @see			Enumeration
     */
    public synchronized Enumeration<V> elements() {
    	if (elementCount == 0) return emptyEnumerator;
    	return new HashIterator(false);
    }
    
    /**
     * Answers a Set of the mappings contained in this Hashtable. Each
     * element in the set is a Map.Entry. The set is backed by this
     * Hashtable so changes to one are relected by the other. The set does
     * not support adding.
     *
     * @return		a Set of the mappings
     */
    public Set<Map.Entry<K, V>> entrySet() {
    	return new Collections.SynchronizedSet(new AbstractSet() {
    		public int size() {return elementCount;}
    		public void clear() {Hashtable.this.clear();}
    		public boolean remove(Object object) {
    			if (contains(object)) {
    				Hashtable.this.remove(((Map.Entry)object).getKey());
    				return true;
    			}
    			return false;
    		}
    		public boolean contains(Object object) {
    			if (object instanceof Map.Entry) {
    				Entry entry = getEntry(((Map.Entry)object).getKey());
    				if (entry == null) {
    					return false;
    				}
    				return entry.equals(object);
    
    			}
    			return false;
    		}
    		public Iterator iterator() {
    			if (elementCount == 0) return emptyIterator;
    			return new HashIterator(MapEntry.Type.entryIterator);
    		}
    	}, this);
    }
    
    /**
     * Compares the specified object to this Hashtable and answer if they
     * are equal. The object must be an instance of Map and contain
     * the same key/value pairs.
     *
     * @param		object	the object to compare with this object
     * @return		true if the specified object is equal to this Map, false otherwise
     *
     * @see			#hashCode
     */
    public synchronized boolean equals(Object object) {
    	if (this == object) return true;
    
    	if (object instanceof Map) {
    		Map map = (Map)object;
    		if (size() != map.size()) return false;
    
    		Set objectSet = map.entrySet();
    		Iterator it = entrySet().iterator();
    		while (it.hasNext())
    			if (!objectSet.contains(it.next())) return false;
    		return true;
    	}
    	return false;
    }
    
    /**
     * Answers the value associated with the specified key in
     * this Hashtable.
     *
     * @param		key	the key of the value returned
     * @return		the value associated with the specified key, null if the specified key
     *				does not exist
     *
     * @see			#put
     */
    public synchronized V get(Object key) {
    	int hash = hash(key); /*ibm@192023*/
    	int index = (hash & 0x7FFFFFFF) % elementData.length;
    	Entry<K, V> entry = elementData[index];
    	while (entry != null) {
    		if (entry.equalsKey(key, hash)) return entry.value;
    		entry = entry.next;
    	}
    	return null;
    }
    
    Entry getEntry(Object key) {
    	int hash = hash(key); /*ibm@192023*/
    	int index = (hash & 0x7FFFFFFF) % elementData.length;
    	Entry entry = elementData[index];
    	while (entry != null) {
    		if (entry.equalsKey(key, hash)) return entry;
    		entry = entry.next;
    	}
    	return null;
    }
    
    /**
     * Answers an integer hash code for the receiver. Objects which are
     * equal answer the same value for this method.
     *
     * @return		the receiver's hash
     *
     * @see			#equals
     */
    public synchronized int hashCode() {
    
    	int result = 0;
    	Iterator it = entrySet().iterator();
    	while (it.hasNext()) {
    		Map.Entry entry = (Map.Entry)it.next();
    		Object key = entry.getKey();
    		Object value = entry.getValue();
    		int hash = (key != this ? key.hashCode() : 0) ^ /*ibm@192925*/
    			(value != this ? (value != null ? value.hashCode() : 0) : 0);
    		result += hash;
    	}
    	return result;
    }
    
    /**
     * Answers if this Hashtable has no key/value pairs, a size of zero.
     *
     * @return		true if this Hashtable has no key/value pairs, false otherwise
     *
     * @see			#size
     */
    public synchronized boolean isEmpty() {
    	return elementCount == 0;
    }
    
    /**
     * Answers an Enumeration on the keys of this Hashtable. The
     * results of the Enumeration may be affected if the contents
     * of this Hashtable are modified.
     *
     * @return		an Enumeration of the keys of this Hashtable
     *
     * @see			#elements
     * @see			#size
     * @see			Enumeration
     */
    public synchronized Enumeration<K> keys() {
    	if (elementCount == 0) return emptyEnumerator;
    	return new HashIterator(true);
    }
    
    /**
     * Answers a Set of the keys contained in this Hashtable. The set is backed
     * by this Hashtable so changes to one are relected by the other. The set does
     * not support adding.
     *
     * @return		a Set of the keys
     */
    public Set<K> keySet() {
    	return new Collections.SynchronizedSet(new AbstractSet() {
    		public boolean contains(Object object) {return containsKey(object);}
    		public int size() {return elementCount;}
    		public void clear() {Hashtable.this.clear();}
    		public boolean remove(Object key) {
    			if (containsKey(key)) {
    				Hashtable.this.remove(key);
    				return true;
    		}
    			return false;
    		}
    		public Iterator iterator() {
    			if (elementCount == 0) return emptyIterator;
    			return new HashIterator(MapEntry.Type.keyIterator);
    		}
    	}, this);
    }
    
    /**
     * Associate the specified value with the specified key in this Hashtable.
     * If the key already exists, the old value is replaced. The key and value
     * cannot be null.
     *
     * @param		key	the key to add
     * @param		value	the value to add
     * @return		the old value associated with the specified key, null if the key did
     *				not exist
     *
     * @see			#elements
     * @see			#get
     * @see			#keys
     * @see			java.lang.Object#equals
     */
    public synchronized V put(K key, V value) {
    	if (key != null && value != null) {
    		int hash = hash(key); /*ibm@192023*/
    		int index = (hash & 0x7FFFFFFF) % elementData.length;
    		Entry<K, V> entry = elementData[index];
    		while (entry != null && !entry.equalsKey(key, hash))
    			entry = entry.next;
    		if (entry == null) {
    			modCount++;
    			if (++elementCount > threshold) {
    				rehash();
					hash = hash(key); /*ibm@192023*/
    				index = (hash & 0x7FFFFFFF) % elementData.length;
    			}
    			if (index < firstSlot) firstSlot = index;
    			if (index > lastSlot) lastSlot = index;
    			entry = new Entry(key, value, hash); /*ibm@192023*/
    			entry.next = elementData[index];
    			elementData[index] = entry;
    			return null;
    		}
    		V result = entry.value;
    		entry.value = value;
    		return result;
    	} else throw new NullPointerException();
    }
    
    /**
     * Copies every mapping in the specified Map to this Hashtable.
     *
     * @param		map the Map to copy mappings from
     */
    public synchronized void putAll(Map<? extends K, ? extends V> map) {
    	Iterator it = map.entrySet().iterator();
    	while (it.hasNext()) {
    		Map.Entry<K, V> entry = (Map.Entry<K, V>)it.next();
    		put(entry.getKey(), entry.getValue());
    	}
    }
    
    /**
     * Increases the capacity of this Hashtable. This method is sent when
     * the size of this Hashtable exceeds the load factor.
     */
    protected void rehash() {
    	int length = (elementData.length<<1)+1;
    	if (length == 0) length = 1;
    	int newFirst = length;
    	int newLast = -1;
    	Entry[] newData = new Entry[length];
		
        /*ibm@192023 Start*/
        boolean currentAltHashing = useAltHashing;
        useAltHashing = sun.misc.VM.isBooted() && 
            (length >= Holder.ALTERNATE_HASHING_THRESHOLD);
        boolean rehash = currentAltHashing ^ useAltHashing;
        /*ibm@192023 End*/
		
    	for (int i=lastSlot+1; --i >= firstSlot;) {
    		Entry entry = elementData[i];
    		while (entry != null) {
                /*ibm@192023 Start*/
                if(rehash) {
                    entry.hashCode = hash(entry.getKey());
                }
                /*ibm@192023 End*/
    			int index = (entry.hashCode & 0x7FFFFFFF) % length;
    			if(index < newFirst) newFirst = index;
    			if(index > newLast) newLast = index;
    			Entry next = entry.next;
    			entry.next = newData[index];
    			newData[index] = entry;
    			entry = next;
    		}
    	}
    	firstSlot = newFirst;
    	lastSlot = newLast;
    	elementData = newData;
    	computeMaxSize();
    }
    /**
     * Remove the key/value pair with the specified key from this Hashtable.
     *
     * @param		key	the key to remove
     * @return		the value associated with the specified key, null if the specified key
     *				did not exist
     *
     * @see			#get
     * @see			#put
     */
    public synchronized V remove(Object key) {
    	int hash = hash(key); /*ibm@192023*/
    	int index = (hash & 0x7FFFFFFF) % elementData.length;
    	Entry<K, V> last = null;
    	Entry<K, V> entry = elementData[index];
    	while (entry != null && !entry.equalsKey(key, hash)) {
    		last = entry;
    		entry = entry.next;
    	}
    	if (entry != null) {
    		modCount++;
    		if (last == null) elementData[index] = entry.next;
    		else last.next = entry.next;
    		elementCount--;
    		V result = entry.value;
    		entry.value = null;
    		return result;
    	}
    	return null;
    }
    /**
     * Answers the number of key/value pairs in this Hashtable.
     *
     * @return		the number of key/value pairs in this Hashtable
     *
     * @see			#elements
     * @see			#keys
     */
    public synchronized int size() {
    	return elementCount;
    }
    
    /**
     * Answers the string representation of this Hashtable.
     *
     * @return		the string representation of this Hashtable
     */
    public synchronized String toString() {
    	if (isEmpty()) return "{}"; //$NON-NLS-1$
    
    	StringBuilder buffer = new StringBuilder(size() * 28);
    	buffer.append('{');
    	for (int i=lastSlot; i >= firstSlot; i--) {
    		Entry entry = elementData[i];
    		while (entry != null) {
    			if (entry.key != this) {
    				buffer.append(entry.key);
    			} else {
    				buffer.append("(this Map)"); //$NON-NLS-1$
    			}
    			buffer.append('=');
    			if (entry.value != this) {
    				buffer.append(entry.value);
    			} else {
    				buffer.append("(this Map)"); //$NON-NLS-1$
    			}
    			buffer.append(", "); //$NON-NLS-1$
    			entry = entry.next;
    		}
    	}
    	// Remove the last ", "
    	if (elementCount > 0) buffer.setLength(buffer.length() - 2);
    	buffer.append('}');
    	return buffer.toString();
    }
    
    /**
     * Answers a Collection of the values contained in this Hashtable. The
     * collection is backed by this Hashtable so changes to one are reflected
     * by the other. The collection does not support adding.
     *
     * @return		a Collection of the values
     */
    public Collection<V> values() {
    	return new Collections.SynchronizedCollection(new AbstractCollection() {
    		public boolean contains(Object object) {return Hashtable.this.contains(object);}
    		public int size() {return elementCount;}
    		public void clear() {Hashtable.this.clear();}
    		public Iterator iterator() {
    			if (elementCount == 0) return emptyIterator;
    			return new HashIterator(MapEntry.Type.valueIterator);
    		}
    	}, this);
    }
    
    private void writeObject(ObjectOutputStream stream) throws IOException {
    	if (CLONE_HASHTABLE_FOR_SYNCHRONIZATION == initialPropValue) {
    		CLONE_HASHTABLE_FOR_SYNCHRONIZATION = (String)AccessController.doPrivileged(new PrivilegedAction() {
    			public Object run() {
    				return System.getProperty("CLONE_HASHTABLE_FOR_SYNCHRONIZATION"); //$NON-NLS-1$
    			}});
    	}
    
    	if (CLONE_HASHTABLE_FOR_SYNCHRONIZATION != null) { //$NON-NLS-1$
    		final Hashtable clone = (Hashtable) this.clone();
    		writeObject(clone, stream);
    		return;
    	}
    
    	synchronized(this) {
    		writeObject(this, stream);
    	}
    }
    
    private void writeObject(Hashtable objToWrite, ObjectOutputStream stream) throws IOException {
    	ObjectOutputStream.PutField fields = stream.putFields();
    	fields.put("threshold", objToWrite.threshold);
    	fields.put("loadFactor", objToWrite.loadFactor);
    	stream.writeFields();
    	stream.writeInt(objToWrite.elementData.length);
    	stream.writeInt(objToWrite.elementCount);
    	for (int i=objToWrite.elementData.length; --i >= 0;) {
    		Entry entry = objToWrite.elementData[i];
    		while (entry != null) {
    			stream.writeObject(entry.key);
    			stream.writeObject(entry.value);
    			entry = entry.next;
    		}
    	}
    }
    private void readObject(ObjectInputStream stream)
    throws IOException, ClassNotFoundException {
    	stream.defaultReadObject();
		
        // set hashMask
        UNSAFE.putIntVolatile(this, HASHMASK_OFFSET, sun.misc.Hashing.randomHashSeed(this)); /*ibm@192023*/
		
    	int length = stream.readInt();
    	elementData = new Entry[length];
    	elementCount = stream.readInt();
        computeMaxSize();                                    /*ibm@192023*/
        useAltHashing = sun.misc.VM.isBooted() &&
            (length >= Holder.ALTERNATE_HASHING_THRESHOLD); /*ibm@192023*/		
    	for (int i=elementCount; --i >= 0;) {
    		Object key = stream.readObject();
    		int hash = hash(key);                          /*ibm@192023*/
    		int index = (hash & 0x7FFFFFFF) % length;
    		if (index < firstSlot) firstSlot = index;
    		if (index > lastSlot) lastSlot = index;
    		Entry entry = new Entry(key, stream.readObject(), hash); /*ibm@192023*/
    		entry.next = elementData[index];
    		elementData[index] = entry;
    	}
    }
}
    